import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np

from smoothmax import Smoothmax

class Exp(nn.Module):
	def __init__(self,maximum=2**6,smoothness=1,inplace=False):
		super(Exp,self).__init__()
		self.inplace = inplace
		self.sm = Smoothmax(maximum,smoothness,inplace=True)
		if maximum==False: self.m = lambda x: x
		else:              self.m = lambda x: self.sm(x)
	def forward(self,x):
		if not self.inplace: x = x.clone()
		if x.is_complex():
			return torch.exp(self.m(x.real)+1j*x.imag)
		else:
			return torch.exp(self.m(x))

class C_Log(nn.Module):
	def __init__(self,minimum=2**-14,smoothness=.5,inplace=False):
		super().__init__()
		self.inplace = inplace
		sm_in  = Smoothmax(       minimum ,-smoothness,inplace=True)
		sm_out = Smoothmax(np.log(minimum),-smoothness,inplace=True)
		if (minimum==False): self.m_in = lambda x: x
		else               : self.m_in = lambda x: torch.sgn(x)*sm_in(x.real.abs())+1j*x.imag
		if (minimum==False): self.m_out = lambda x: x
		else               : self.m_out = lambda x: sm_out(x.real)+1j*x.imag
	def forward(self,x):
		if not self.inplace: x = x.clone()
		x = x.type(torch.cfloat)
		### input capping :
		# x = self.m_in(x)
		### compute log :
		x = torch.log(x)
		### output capping :
		x = self.m_out(x)
		return x

###############################################################################
### RELU ######################################################################
###############################################################################

class C_ReLU(nn.Module):
	def __init__(self,mode="simple",inplace=False,relu=F.relu):
		super().__init__()
		self.inplace = inplace
		self.relu = relu
	def forward(self,x):
		if not self.inplace: x = x.clone()
		x = x.type(torch.cfloat)
		if   mode=="simple": return self.relu(x.real)+1j*x.imag
		elif mode=="cos"   : return x*torch.cos(x.angle()/2)
		elif mode=="harsh" : return x*self.relu(x.real)/x.real # breaks continuity
		# elif mode=="harsh" : return x*(x.real>0) # breaks continuity
		### overkill
		# return F.relu(x.real) + 1j*x.imag*torch.exp(torch.minimum(x.real,torch.zeros(x.shape)))
		### broken
		# return x*torch.cos(x.angle())                    # totally broken
		# return F.relu(x.abs()-1)*torch.exp(1j*x.angle()) # totally broken

###############################################################################
### Basic #####################################################################
###############################################################################

class MSELoss(nn.Module):
	def __init__(self):
		super().__init__()
	def forward(self,x,y):
		e=x-y
		return ((e*torch.conj(e)).real).mean()

# idea from https://discuss.pytorch.org/t/how-can-i-dropout-for-torch-complex-data-type/135577
# ! this should do the scaling of gradient from dropout
class Dropout(nn.Module):
	def __init__(self,p=0.5):
		super(Dropout,self).__init__()
		# self.do=nn.Dropout(p)
		self.p=p
	def forward(self,x):
		mask = torch.nn.functional.dropout(torch.ones_like(x.real),self.p,self.training)
		return x * mask

###############################################################################
### CONVERSION ################################################################
###############################################################################

class ToComplex(nn.Module):
	def __init__(self): super(ToComplex,self).__init__()
	def forward(self,x): return x+0j

class ToReal(nn.Module):
	def __init__(self): super(ToReal,self).__init__()
	def forward(self,x): return x.real


